/*  Copyright (C) 2002  Byron Stanoszek
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **************************************************************************/

/* mareport.c */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <netdb.h>
#include <errno.h>
#include <arpa/inet.h>
#include <arpa/telnet.h>

#define ANSWERING_MACHINE    "login.txt"
#define TOO_MANY_CONNECTIONS "toomany.txt"
#define SERVER_SCRIPT        "script.txt"

#define DEFAULT_PORT 7348

#define VERSION "2.1.1"


/* Descriptor Flags */

/* Socket Types */
#define CLIENT		0x1	// Active client connection
#define REMOTE_SOCKET	0x2	// Socket to remote TinyMare server
#define IDENT_SOCKET	0x4	// Non-blocking ident authentication

/* Client Flags */
#define CONNECTED	0x10	// Client is fully connected to the server
#define DNS_WAIT	0x20	// Waiting for DNS to resolve hostname
#define IDENT_WAIT	0x40	// Waiting on user identification
#define LOGOFF		0x80	// Connection closed pending on output flush

/* Socket Flags */
#define CONNECT_WAIT	0x100	// Waiting for connect() to establish
#define HANDSHAKE_WAIT	0x200	// Waiting for server to send protocol string
#define ESTABLISHED	0x400	// Successfully connected to remote socket


#define debug(x...) ({ if(debug) printf(x); })

#define ipaddr(x) ({ struct in_addr _addr; _addr.s_addr=x; inet_ntoa(_addr); })

#define MAX_DATA_QUEUE	1048576  // Size of output queue to TinyMare server
#define MAX_OUTPUT	131072   // Size of output queue per client

#define IO_CHUNK_SIZE   (8192-sizeof(struct io_queue))

struct conc_input {
  int offset;
  int concid;
  unsigned short len;
  unsigned char *data;
};

struct io_queue {
  struct io_queue *next;
  int len;			// Number of bytes already written to buffer
  int offset;			// Number of bytes already read from buffer
  unsigned char data[0];	// Buffer (usually of size IO_CHUNK_SIZE)
};

typedef struct descriptor_list {
  struct descriptor_list *next;
  int fd;			// FD of client socket connection

  int flags;			// State of descriptor entry
  int concid;			// Remote Connection ID
  int local_id;			// Local ID
  int addr;			// IP Address
  unsigned short port;		// IP Port
  char hostname[65];		// Hostname

  struct io_queue *output;	// Output buffer
  int output_len;		// Total amount of output pending

  unsigned char input[8192];	// Current input line (from client)
  int input_offset;		// Number of bytes written to input line

  time_t timeout;		// Timeout on user identification
} DESC;

DESC *descriptor_list;

int sock=-1;		// Local Port Socket
int resolv=-1;		// Resolver Socket

int top_desc;		// Maximum FD to select()
int local_id;		// Next Local ID to issue
int debug;		// 1=Debugging messages sent to Standard Output

DESC *remote_desc;	// Pointer to remote server descriptor
time_t remote_timer;	// When to try the remote connection again

char localhost[128];
char *remote_host;

int remote_addr=INADDR_NONE;
int port=DEFAULT_PORT;
int remote_port;

char mudname[65];
int num_users;		// Current number of connections active
int open_files;		// Number of open file descriptors set by system
int limit;		// Limit number of possible logins

struct conc_input conc_buf;  // Current input buffer for concentrator

time_t now, lastping, ping_timer;


/* function declarations */

extern char *dns_lookup(int ip);
extern int  dns_open(void);
extern void dns_receive(void);
extern void dns_events(void);

static int  make_socket(int port);
static DESC *initializesock(int s, int type);
static void serverport(void);
static void mare_connect(void);
static void new_connection(void);
static void init_connect(DESC *d);
static void close_server(void);
static void close_connections(DESC *k);

static void check_connect(DESC *d);
static void queue_output(DESC *d, void *msg, int len);
static void send_file(DESC *d, char *filename);

static int  process_output(DESC *d);
static void protocol_handshake(void);
static void client_connect(DESC *d);
static void client_disconnect(int concid);
static void send_data(int concid, char *buf, int len);
static int  process_input(DESC *d);
static void process_input_internal(DESC *d, unsigned char *buf, int len);
static int  telnet_iac(DESC *d, unsigned char *s, int len);
static int  service_conc(char *msg, int len);
static int  process_conc_command(struct conc_input *io);

static void connect_ident(DESC *d);
static void send_ident_request(DESC *d);
static int  process_ident(DESC *d, char *input, int size);

#ifdef __CYGWIN__
extern int getdomainname(char *name, size_t len);
#endif

// Program entry: Process arguments, open socket, start server.
//
int main(int argc, char *argv[])
{
  struct hostent *hent;
  char hostname[64], domainname[64];
  int c;

  while((c = getopt(argc, argv, "dl:p:v")) != -1) {
    switch(c) {
    case 'd':  /* Increment debug level */
      debug++;
      break;
    case 'l':  /* Limit number of users */
      limit=atoi(optarg);
      if(limit < 0)
        limit=0;
      break;
    case 'p':  /* Select local port */
      port=atoi(optarg);
      if(port < 1 || port > 65535) {
        fprintf(stderr, "%s: Invalid port number %d\n", *argv, port);
        exit(1);
      }
      break;
    case 'v':
      printf("Mareport version %s\n", VERSION);
      exit(0);
    case ':':
      fprintf(stderr, "%s: option requires an argument -- %c\n",*argv,optopt);
      exit(1);
    case '?':
      fprintf(stderr, "Usage: %s [-dv] [-l limit] [-p localport] remotehost "
              "[remoteport]\n", *argv);
      exit(1);
    }
  }

  /* Check port number */
  if(port < 1024 && geteuid()) {
    fprintf(stderr, "%s: Only root can run on port %d\n", *argv, port);
    exit(1);
  }

  /* Check for a remote host */
  if((remote_host=argv[optind++])) {
    if(argv[optind]) {
      remote_port=atoi(argv[optind]);
      if(remote_port < 1 || remote_port > 65535) {
        fprintf(stderr, "%s: Invalid port number %d\n", *argv, remote_port);
        exit(1);
      }
    } else
      remote_port=DEFAULT_PORT;
  }

  signal(SIGPIPE, SIG_IGN);
  signal(SIGHUP, SIG_IGN);

#ifndef __CYGWIN__
  /* Open DNS Resolver */
  resolv=dns_open();
#endif

  /* Open socket on local port */
  top_desc=sock=make_socket(port);

  /* Determine local hostname */
  if(gethostname(hostname, 60) == -1 || getdomainname(domainname, 60) == -1)
    strcpy(localhost, "concentrator");
  else {
    char *s=domainname;

    strcpy(localhost, hostname);
    while(*s == '.')
      s++;
    if(*s && strcmp(s, "(none)"))
      sprintf(localhost+strlen(localhost), ".%s", s);
  }

  /* Resolve remote hostname */
  if(remote_host) {
    if(!(hent=gethostbyname(remote_host))) {
      herror(remote_host);
      exit(1);
    }
    remote_addr = *(long *)hent->h_addr_list[0];
  }

  printf("%s online at port %d.\n", remote_host?"Concentrator service":
         "Answering machine", port);

  if(!debug) {
    /* Fork into the background */
    setsid();
    if((c=fork()) == -1) {
      perror("fork()");
      exit(1);
    } if(c)
      exit(0);

    fclose(stdin);
    fclose(stdout);
    fclose(stderr);
  } else {
    setvbuf(stdout, NULL, _IONBF, 0);
    fclose(stdin);
    fclose(stderr);
  }

  /* determine max number of open files allowed */
#ifdef __CYGWIN__
  open_files=-1;
#elif defined(RLIMIT_NOFILE)
  {
    struct rlimit rlim;
    getrlimit(RLIMIT_NOFILE, &rlim);
    open_files=rlim.rlim_cur;
  }
#else
  open_files=sysconf(_SC_OPEN_MAX);
#endif

  if((unsigned int)open_files > 0x7ffffffe)
    open_files=-1;
  if(open_files > 4) {
    open_files-=4;
    open_files-=open_files % 5;
  }

  /* Start Server */
  serverport();

  /* Exit Application */
  exit(0);
}

// Opens a socket on <port> to listen for new connections with a
// back queue of 128. Returns the FD of the socket opened.
//
static int make_socket(int port)
{
  struct sockaddr_in server;
  char buf[32];
  int s, on=1;

  sprintf(buf, "port %d", port);

  /* Create socket */
  if((s=socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    perror(buf);
    exit(1);
  }

  /* Allow port to be reused in the future */
  if(setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on)) < 0) {
    perror(buf);
    close(s);
    exit(1);
  }

  /* Make socket nonblocking for accept() */
  fcntl(s, F_SETFL, O_NONBLOCK);

  /* Bind the socket to an address:port */
  server.sin_family = AF_INET;
  server.sin_addr.s_addr = INADDR_ANY;
  server.sin_port = htons(port);
  if(bind(s, (struct sockaddr *)&server, sizeof(server)) < 0) {
    perror(buf);
    close(s);
    exit(1);
  }

  /* Listen for connections */
  listen(s, 128);

  return s;
}

/* Initialize a descriptor */
static DESC *initializesock(int s, int type)
{
  DESC *d;

  if(s > top_desc)
    top_desc=s;

  /* Make connection non-blocking */
  fcntl(s, F_SETFL, O_NONBLOCK);

  /* Allocate a new descriptor structure and zero its contents */
  d=malloc(sizeof(DESC));
  d->next=descriptor_list;
  descriptor_list=d;

  /* Initialize data on descriptor pointer */
  d->fd=s;
  d->flags=type;
  d->concid=0;
  d->local_id=0;
  d->output=NULL;
  d->output_len=0;
  d->input_offset=0;
  d->timeout=0;

  *d->hostname=0;

  return d;
}

// Main Loop: Process server requests, new connections, data forwarding.
//
static void serverport()
{
  DESC *d, *next;
  struct timeval timeout;
  static fd_set *input, *output;
  static int fdset_size;
  int found;

  now=lastping=time(NULL);

  while(1) {
    /* Manage non-blocking connection to remote server */
    if(remote_host && !remote_desc && now >= remote_timer)
      mare_connect();

    /* Dynamically grow file descriptor sets when needed (for >1024 users) */
    if(!input || fdset_size*8 <= top_desc) {
      free(input);
      free(output);
      fdset_size=(top_desc/64+1)*8;
      input=malloc(fdset_size);
      output=malloc(fdset_size);
    }

    /* Clear input/output file descriptor sets */
    memset(input, 0, fdset_size);
    memset(output, 0, fdset_size);

    /* Wake up when we get a new connection */
    FD_SET(sock, input);

    /* Manage DNS resolver socket */
    if(resolv != -1)
      FD_SET(resolv, input);

    /* Go through descriptor list and add all we need for select() */
    for(d=descriptor_list;d;d=d->next) {
      if((d->flags & CONNECT_WAIT) || d->output)
        FD_SET(d->fd, output);  // Output pending, or waiting on connect()
      if(!(d->flags & CONNECT_WAIT))
        FD_SET(d->fd, input);
    }

    timeout.tv_sec=1;
    timeout.tv_usec=0;

    /* Wait with 1 second timeout for DNS Resolver & Ping processing */
    found=select(top_desc+1, input, output, NULL, &timeout);

    now=time(NULL);

#ifndef __CYGWIN__
    if(resolv != -1) {
      static time_t last_check;

      /* Check for resolver responses */
      if(FD_ISSET(resolv, input))
        dns_receive();

      /* Call DNS routines once per second */
      if(last_check != now) {
        last_check=now;

        /* Update DNS resolver cache */
        dns_events();

        /* Check to see if the IP Address has been resolved */
        for(d=descriptor_list;d;d=d->next)
          if(d->flags & DNS_WAIT)
            init_connect(d);
      }
    }
#endif

    /* Notify players every 5 minutes when the game hangs */
    if(remote_desc && (remote_desc->flags & ESTABLISHED) &&
       (now-lastping)/60 != ping_timer) {
      ping_timer=(now-lastping)/60;
      if(!(ping_timer % 5)) {
        char buf[256];

        sprintf(buf,"%% The connection to %s has been idle for %ld minutes.\n",
                *mudname?mudname:"the mare", ping_timer);
        for(d=descriptor_list;d;d=d->next)
          if(d->flags & CONNECTED)
            queue_output(d, buf, strlen(buf));
      }
    }

    if(found) {
      /* Process input FDs */
      for(d=descriptor_list;d;d=next) {
        next=d->next;
        if(!FD_ISSET(d->fd, input))
          continue;

        /* Check for reading handshake data from the server */
        if(d->flags & HANDSHAKE_WAIT)
          protocol_handshake();
        else if(!process_input(d)) {
          if(d->flags & REMOTE_SOCKET)
            close_server();
          else {
            if(d->flags & CONNECTED)
              client_disconnect(d->concid);
            close_connections(d);
          }
        }
      }

      /* Process output FDs */
      for(d=descriptor_list;d;d=next) {
        next=d->next;
        if(!FD_ISSET(d->fd, output))
          continue;

        /* Check if the nonblocking connect() has succeeded */
        if(d->flags & CONNECT_WAIT)
          check_connect(d);
        else if(d->output && !process_output(d)) {
          if(d->flags & REMOTE_SOCKET)
            close_server();
          else {
            if(d->flags & CONNECTED)
              client_disconnect(d->concid);
            close_connections(d);
          }
        }
      }

      /* Accept a new connection on the local port */
      if(FD_ISSET(sock, input))
        new_connection();
    }

    /* Close expired connections */
    close_connections(NULL);
  }
}

// Attempt to start connection to the remote TinyMARE server.
//
static void mare_connect()
{
  struct sockaddr_in addr;
  int s, ret;

  /* Open new socket for remote connection */
  if((s=socket(AF_INET, SOCK_STREAM, 0)) == -1)
    return;

  fcntl(s, F_SETFL, O_NDELAY);

  addr.sin_family=AF_INET;
  addr.sin_port=htons(remote_port);
  addr.sin_addr.s_addr=remote_addr;

  /* Try nonblocking connect */
  if((ret=connect(s, (struct sockaddr *)&addr, sizeof(addr))) == -1 &&
     errno != EINPROGRESS) {
    close(s);
    remote_timer=now+5;
    debug("Connection to mare failed, waiting 5 seconds...\n");
    return;
  }

  /* Initialize descriptor for remote connection */
  remote_desc=initializesock(s, REMOTE_SOCKET);

  if(ret == -1) {  /* connect cannot complete immediately */
    remote_desc->flags |= CONNECT_WAIT;
    debug("Connection to mare in progress.\n");
  } else {
    remote_desc->flags |= HANDSHAKE_WAIT;
    debug("Connection to mare established.\n");
    send_file(remote_desc, SERVER_SCRIPT);
  }
}

// Check if the non-blocking connection succeeded.
//
static void check_connect(DESC *d)
{
  int err, err_size=sizeof(int);
  DESC *k;

  /* Check error on non-blocking connect */
  getsockopt(d->fd, SOL_SOCKET, SO_ERROR, &err, &err_size);

  /* Check for connection to remote TinyMare server */
  if(d == remote_desc) {
    if(err) {
      d->flags=LOGOFF;
      remote_desc=NULL;
      remote_timer=now+5;
      debug("Nonblocking connection to mare failed, waiting 5 seconds...\n");
      return;
    }

    d->flags=REMOTE_SOCKET|HANDSHAKE_WAIT;
    debug("Nonblocking connection to mare established.\n");
    send_file(d, SERVER_SCRIPT);
    return;
  }

  /* Otherwise this is a connection to an identd service */

  /* Scan for matching client */
  for(k=descriptor_list;k;k=k->next) {
    if((k->flags & CLIENT) && k->local_id == d->local_id)
      break;
  }
  if(!k) {
    d->flags|=LOGOFF;
    return;
  }

  if(err) {
    d->flags|=LOGOFF;
    debug("Nonblocking ident lookup failed for FD %d\n", k->fd);
    return;
  }

  d->flags=IDENT_SOCKET|ESTABLISHED;
  send_ident_request(d);
}

// Accept a new connection from the local port.
//
void new_connection()
{
  DESC *d;
  struct sockaddr_in addr;
  int fd, lim, addr_len=sizeof(struct sockaddr);
  char buf[1024], *s;

  if((fd=accept(sock, (struct sockaddr *)&addr, &addr_len)) < 0)
    return;

  /* add new connection entry */
  d=initializesock(fd, CLIENT);

  d->addr=addr.sin_addr.s_addr;
  d->port=ntohs(addr.sin_port);

  d->local_id=local_id++;

  debug("Accepted connection as FD %d from %s\n", d->fd, ipaddr(d->addr));

  num_users++;
  lim=(limit > 0 && limit < open_files)?limit:open_files;

  /* display login information */
  if(!remote_host || !remote_desc || !(remote_desc->flags & ESTABLISHED) ||
     now-lastping >= 300 || (lim > 0 && num_users > lim)) {

    s=buf+sprintf(buf, "\nWelcome to network system: %s %d.\nLocal time is %s",
                  localhost, port, ctime(&now));
    if(lim < 1 || num_users-1 > lim)
      sprintf(s, "There %s %d user%s online.\n\n", num_users==1?"is":"are",
              num_users, num_users==1?"":"s");
    else
      sprintf(s, "There are %d/%d users online.\n\n", (num_users > lim)?lim:
              num_users, lim);
    queue_output(d, buf, strlen(buf));

    if(lim > 0 && num_users > lim) {
      send_file(d, TOO_MANY_CONNECTIONS);
      d->flags=CLIENT|LOGOFF;
      return;
    }

    send_file(d, ANSWERING_MACHINE);
    if(!remote_host) {
      d->flags=CLIENT|LOGOFF;
      return;
    }

    if(!remote_desc || !(remote_desc->flags & ESTABLISHED))
      s=buf+sprintf(buf, "%s has been down for ", *mudname?mudname:"The mare");
    else
      s=buf+sprintf(buf, "The connection to %s has been idle for ", mudname);

    if(now-lastping >= 86400)
      s+=sprintf(s, "%ld day%s.\n", (now-lastping)/86400,
                 (now-lastping)/86400==1?"":"s");
    else if(now-lastping >= 3600)
      s+=sprintf(s, "%ld hour%s.\n", (now-lastping)/3600,
                 (now-lastping)/3600==1?"":"s");
    else
      s+=sprintf(s, "%ld minute%s.\n", (now-lastping)/60,
                 (now-lastping)/60==1?"":"s");

    strcpy(s, "Please wait...\n\n");
    queue_output(d, buf, strlen(buf));
  }

  /* Attempt to lookup hostname */
  if(!d->addr || ntohl(d->addr) >> 24 == 127)
    strcpy(d->hostname, localhost);
  else if((s=dns_lookup(d->addr))) {
    sprintf(d->hostname, "%.64s", s);
    debug("Host lookup succeeded for FD %d: %s\n", d->fd, s);
  } else
    d->flags|=DNS_WAIT;

  /* Create non-blocking connection to remote identd server */
  connect_ident(d);

  if(!(d->flags & (DNS_WAIT|IDENT_WAIT)))
    init_connect(d);
}

static void init_connect(DESC *d)
{
  char *s;

  /* Check if hostname has been resolved */
  if((d->flags & DNS_WAIT) && (s=dns_lookup(d->addr))) {
    if(!*d->hostname)
      sprintf(d->hostname, "%.64s", s);
    else
      sprintf(d->hostname+strlen(d->hostname), "@%.*s",
              63-(int)strlen(d->hostname), s);

    d->flags &= ~DNS_WAIT;
    debug("Host lookup succeeded for FD %d: %s\n", d->fd, s);
  }

  /* Check if all is resolved.. */
  if(!(d->flags & (DNS_WAIT|IDENT_WAIT)))
    if(remote_desc && (remote_desc->flags & ESTABLISHED))
      client_connect(d);  // Success! Send client connect string to TinyMARE.
}

static void close_server()
{
  DESC *d;

  debug("Lost connection to %s.\n", *mudname?mudname:"remote server");
  remote_desc->flags=LOGOFF;
  remote_desc=NULL;
  remote_timer=now+5;

  /* Close established connections */
  for(d=descriptor_list;d;d=d->next)
    if(d->flags & CONNECTED)
      d->flags=CLIENT|LOGOFF;

  /* Clear concentrator input buffer */
  free(conc_buf.data);
  conc_buf.offset=0;
  conc_buf.data=NULL;
}

// Close clients connected to the concentrator service.
//
// If <k> is not specified, close all clients whose states are LOGOFF and have
// no output pending.
//
static void close_connections(DESC *k)
{
  DESC *d, *ptr, *last=NULL, *next;
  struct io_queue *io, *next_io;

  for(d=descriptor_list;d;d=next) {
    next=d->next;
    if((k && (d != k)) ||
       (!k && (d->output || !((d->flags & LOGOFF) ||
        (d->timeout && now >= d->timeout))))) {
      last=d;
      continue;
    }

    debug("FD %d closed: %s\n", d->fd, (d->flags & IDENT_SOCKET)?
          "Ident connection closed.":k?"Client connection exited.":
          "Requested by server.");

    if(d->flags & (CLIENT|IDENT_SOCKET))
      num_users--;

    if(d->flags & IDENT_SOCKET) {
      /* Match client descriptor */
      for(ptr=descriptor_list;ptr;ptr=ptr->next)
        if((ptr->flags & CLIENT) && (ptr->flags & IDENT_WAIT) &&
           ptr->local_id == d->local_id) {
          ptr->flags &= ~IDENT_WAIT;
          init_connect(ptr);
          break;
        }
    }

    close(d->fd);
    for(io=d->output;io;io=next_io) {
      next_io=io->next;
      free(io);
    }
    free(d);
    if(last)
      last->next=next;
    else
      descriptor_list=next;

    if(k)
      break;
  }
}

// Append a data buffer to an output queue.
//
static void queue_output(DESC *d, void *msg, int len)
{
  struct io_queue *new, *last;
  int size, offset=0;

  if(!d)
    return;

  for(last=d->output;last && last->next;last=last->next);

  while(offset < len) {
    if(!last || last->len >= IO_CHUNK_SIZE) {
      new=malloc(sizeof(struct io_queue)+IO_CHUNK_SIZE);
      new->next=NULL;
      new->len=new->offset=0;

      if(last)
        last->next=new;
      else
        d->output=new;
      last=new;
    }
    size=IO_CHUNK_SIZE - last->len;
    size=(len-offset < size) ? (len-offset) : size;
    memcpy(last->data+last->len, (char *)msg+offset, size);
    last->len += size;
    offset += size;
  }

  d->output_len += len;
}

// Append the contents of a file to an output queue.
//
static void send_file(DESC *d, char *filename)
{
  char buf[4096];
  int fd, len;

  if((fd=open(filename, O_RDONLY)) != -1) {
    while((len=read(fd, buf, 4096)) > 0)
      queue_output(d, buf, len);
    close(fd);
  }
}

// Wait for a specific message to arrive from the server.
//
void protocol_handshake()
{
  DESC *d;
  static char buf[1024];
  static int offset;
  char *r=buf, *s, *t;
  int a, num, len;

  if((len=read(remote_desc->fd, buf+offset, 1024-offset)) <= 0) {
    remote_desc->flags=LOGOFF;
    remote_desc=NULL;
    remote_timer=now+300;

    debug("Remote server closed connection before protocol handshaking "
          "was completed.\n");
    offset=0;
    return;
  }

  offset+=len;

  while(r < buf+1024) {
    /* Check if a newline is received */
    if(!(s=memchr(r, '\n', offset-(r-buf)))) {
      if(offset == 1024 && r == buf) {
        offset=0;
        return;
      }
      break;
    }
    t=s+1;

    if(s > r && *(s-1) == '\r')
      s--;
    *s='\0';
    if(debug > 1)
      debug("Line received: '%s'\n", r);
    if(strncmp(r, "Concentrator Protocol ", 22)) {
      r=t;
      continue;
    }
    offset=0;  /* clear buffer offset for next time */
    r+=22;
    if((s=strchr(r, ' ')) && *(s+1))
      sprintf(mudname, "%.64s", s+1);
    else
      strcpy(mudname, "TinyMARE");
    if(s)
      *s='\0';
    if(strcmp(r, "2.0")) {
      debug("Remote server uses unsupported concentrator protocol '%s'.\n", r);
      remote_desc->flags=LOGOFF;
      remote_desc=NULL;
      remote_timer=now+300;
      return;
    }

    debug("Established concentrator protocol %s. Remote host: %s\n",
          r, mudname);

    remote_desc->flags=REMOTE_SOCKET|ESTABLISHED;
    lastping=now;
    ping_timer=0;

    /* Connect all clients in reverse order */
    for(num=0,d=descriptor_list;d;d=d->next,num++);
    for(;num > 0;num--) {
      for(a=1,d=descriptor_list;a < num;d=d->next,a++);
      if(d->flags == CLIENT)  /* and not waiting for DNS or ident.. */
        client_connect(d);
    }
    return;
  }

  if(r != buf) {
    memmove(buf, r, offset-(r-buf));
    offset-=(r-buf);
  }
}

// Create a concentrator login on the TinyMARE server.
//
static void client_connect(DESC *d)
{
  struct {
    int concid;
    unsigned short len;
    char cmd;
    int local_id;
    int addr;
    short port;
    char hostname[65];
  } __attribute__((__packed__)) buf;

  int len=strlen(d->hostname);

  buf.concid=0;
  buf.len=htons(len+12);
  buf.cmd=1;
  buf.local_id=htonl(d->local_id);
  buf.addr=d->addr;  /* Already big-endian */
  buf.port=htons(d->port);
  memcpy(buf.hostname, d->hostname, len+1);

  queue_output(remote_desc, &buf, len+18);
}

// Disconnect a client from the remote TinyMARE server.
//
static void client_disconnect(int concid)
{
  struct {
    int remote_concid;
    unsigned short len;
    char cmd;
    int concid;
  } __attribute__((__packed__)) buf={0, htons(5), 2, htonl(concid)};

  queue_output(remote_desc, &buf, sizeof(buf));
}

// Send text data to the server using a specific Connection ID.
//
static void send_data(int concid, char *buf, int len)
{
  struct {
    int concid;
    unsigned short len;
  } __attribute__((__packed__)) header={htonl(concid), htons(len)};

  if(!remote_desc || remote_desc->output_len > MAX_DATA_QUEUE)
    return;

  queue_output(remote_desc, &header, sizeof(header));
  queue_output(remote_desc, buf, len);
}

// Sends pending output to a file descriptor.
//
static int process_output(DESC *d)
{
  char buffer[32768];
  struct io_queue *next, *new;
  int amt, len, size=0;

  /* Construct a giant buffer (up to 32768 bytes) to do an atomic send over
     the network. Any data left over gets tacked onto the beginning of the
     server output queue for the next successful select(). */

  while(d->output && size < 32768) {
    next=d->output->next;
    amt=d->output->len - d->output->offset;
    if(amt > 32768-size)
      amt=32768-size;
    memcpy(buffer+size, d->output->data + d->output->offset, amt);
    size+=amt;

    if((d->output->offset += amt) == d->output->len) {
      free(d->output);
      d->output=next;
    }
  }

  /* Send data buffer */
  if((len=write(d->fd, buffer, size)) < size) {
    if(len == -1) {
      if(errno != EINTR)
        return 0;  /* Connection has unexpectedly closed */
      len=0;
    }

    /* Move unsent data to front of output queue */
    new=malloc(sizeof(struct io_queue)+((!d->output &&
               size-len < IO_CHUNK_SIZE)?IO_CHUNK_SIZE:(size-len)));
    new->next=d->output;
    new->len=size-len;
    new->offset=0;
    memcpy(new->data, buffer+len, size-len);
    d->output=new;
  }

  d->output_len -= len;
  return 1;
}

// Process input from all established connections.
//
static int process_input(DESC *d)
{
  unsigned char buf[4096];
  int len;

  /* Read data & check that socket is still alive */
  if((len = read(d->fd, buf, 4096)) <= 0) {
    if(len == -1 && errno == EINTR)
      return -1;
    return 0;  /* Connection has unexpectedly closed */
  }

  if(d == remote_desc)
    return service_conc(buf, len);
  if(d->flags & IDENT_SOCKET)
    return process_ident(d, buf, len);

  /* If the connection is not yet established, buffer up to 8000 bytes */
  if(!(d->flags & CONNECTED)) {
    if(d->input_offset + len > 8000)
      len=8000 - d->input_offset;
    memcpy(d->input+d->input_offset, buf, len);
    d->input_offset+=len;
    return 1;
  }

  process_input_internal(d, buf, len);
  return 1;
}

static void process_input_internal(DESC *d, unsigned char *buf, int len)
{
  int a;

  /* Filter input */
  for(a=0;a<len;a++) {
    /* Process telnet command characters */
    if(buf[a] == IAC && a < len-1) {
      a+=telnet_iac(d, buf+a, len-a);
      continue;
    }

    if(buf[a] == 8 || buf[a] == 127) {  /* backspace */
      if(d->input_offset > 0)
        d->input_offset--;
      continue;
    }
    if(buf[a] == 10) {  /* newline */
      d->input[d->input_offset++]='\n';
      send_data(d->concid, d->input, d->input_offset);
      d->input_offset=0;
      continue;
    }
    if(buf[a] >= 32 && buf[a] < 127 && d->input_offset < 8000)
      d->input[d->input_offset++]=buf[a];
  }
}

/* Thin-client Telnet "Is A Command" Character Processing */
static int telnet_iac(DESC *d, unsigned char *s, int len)
{
  int i;

  /* Pass telnet sequence to remote host immediately. But first we need to
     determine the complete size of the sequence (mainly for SB packets) */

  /* Handle Subnegotiation (SB) */
  if(len > 2 && s[1] == SB) {
    for(i=2;i<len;i++)
      if(s[i] == IAC && i < len-1 && s[i+1] == SE) {
        i+=2;
        break;
      }
  } else
    i=(*s < 251)?2:3;

  send_data(d->concid, s, i);
  return i-1;
}

// Process concentrator input from the remote TinyMARE server.
//
int service_conc(char *msg, int len)
{
  struct conc_input *io=&conc_buf;
  int result, size, offset=0;

  while(offset < len) {
    /* Read message header */
    if(!io->data) {
      if((size=len-offset) > 6-io->offset)
        size=6-io->offset;
      memcpy((char *)&io->concid+io->offset, msg+offset, size);
      io->offset += size;
      if(io->offset < 6)
        return 1;  /* Need more data for header */

      offset += size;
      io->offset=0;
      io->concid=ntohl(io->concid);
      io->len=ntohs(io->len);
      io->data=malloc(io->len ?: 1);
    }

    /* Read message contents */
    if((size=len-offset) > io->len-io->offset)
      size=io->len-io->offset;
    memcpy(io->data+io->offset, msg+offset, size);
    offset += size;
    io->offset += size;
    if(io->offset == io->len) {
      result = process_conc_command(io);
      free(io->data);
      io->data=NULL;
      io->offset=0;
      if(!result)
        return 0;
    }
  }

  return 1;
}

#define get16(p) (p+=2, (p[-2]<<8) | p[-1])
#define get32(p) (p+=4, (p[-4]<<24) | (p[-3]<<16) | (p[-2]<<8) | p[-1])

int process_conc_command(struct conc_input *io)
{
  DESC *d;
  int local_id, concid, temp;
  unsigned char *s;

  /* Process concentrator command */
  if(!io->concid) {
    if(!io->len)
      return 0;

    switch(io->data[0]) {
    case 2:  // Disconnect Client
      if(io->len < 5)
        return 0;

      /* Copy big-endian integer arguments */
      s=io->data+1;
      concid=get32(s);

      for(d=descriptor_list;d;d=d->next)
        if((d->flags & CONNECTED) && d->concid == concid) {
          d->flags=CLIENT|LOGOFF;
          break;
        }
      return 1;

    case 3:  // Identity Response
      if(io->len < 9)
        return 0;
      
      /* Copy big-endian integer arguments */
      s=io->data+1;
      local_id=get32(s);
      concid=get32(s);

      /* Match local_id with descriptor */
      for(d=descriptor_list;d;d=d->next)
        if(d->flags == CLIENT && d->local_id == local_id) {
          d->concid=concid;
          d->flags|=CONNECTED;

          /* Send buffered input to server */
          if((temp=d->input_offset)) {
            d->input_offset=0;
            process_input_internal(d, d->input, temp);
          }
          break;
        }

      /* If descriptor wasn't found, send disconnect message */
      if(!d)
        client_disconnect(concid);
      return 1;

    case 4:  // Ping
      lastping=now;
      ping_timer=0;
      return 1;
    }

    /* Invalid Opcode; Close connection */
    return 0; 
  }

  /* Determine which client we're talking to */
  for(d=descriptor_list;d;d=d->next)
    if((d->flags & CONNECTED) && d->concid == io->concid)
      break;

  /* Queue data in client's output buffer */
  if(d && d->output_len < MAX_OUTPUT)
    queue_output(d, io->data, io->len);

  return 1;
}

// Open a non-blocking socket to query the remote host's identity.
//
static void connect_ident(DESC *d)
{
  struct sockaddr_in localaddr, sin_addr;
  int len, fd, ret;
  DESC *k;

  len = sizeof(localaddr);
  if(getsockname(sock, (struct sockaddr *)&localaddr, &len))
    return;

  /* Open non-blocking socket */
  if((fd=socket(AF_INET, SOCK_STREAM, 0)) == -1)
    return;
  fcntl(fd, F_SETFL, O_NDELAY);

  sin_addr.sin_family=AF_INET;
  sin_addr.sin_addr=localaddr.sin_addr;
  sin_addr.sin_port=0;
  if(bind(fd, (struct sockaddr *)&sin_addr, sizeof(sin_addr))) {
    close(fd);
    return;
  }

  sin_addr.sin_family = AF_INET;
  sin_addr.sin_addr.s_addr = d->addr;
  sin_addr.sin_port = htons(113); /* identd port */

  /* Attempt a non-blocking connect() to the remote address */
  if((ret=connect(fd, (struct sockaddr *)&sin_addr, sizeof(sin_addr))) == -1 &&
     errno != EINPROGRESS) {
    close(fd);
    debug("Connection to identd failed for FD %d\n", d->fd);
    return;
  }

  /* Tell main session to wait for identification */
  d->flags |= IDENT_WAIT;

  /* Initialize descriptor for ident authentication */
  k=initializesock(fd, IDENT_SOCKET);
  k->local_id=d->local_id;
  k->port=d->port;

  num_users++;

  if(ret == -1) {  /* connect cannot complete immediately */
    k->flags |= CONNECT_WAIT;
    k->timeout=now+3;
  } else {
    k->flags |= ESTABLISHED;
    send_ident_request(k);
  }
}

// Send identification request to remote identd server.
//
static void send_ident_request(DESC *d)
{
  char buf[32];

  sprintf(buf, "%d,%d\r\n", d->port, port);
  write(d->fd, buf, strlen(buf));

  d->timeout=now+3;
}

// Process input on an ident-connected descriptor.
//
static int process_ident(DESC *d, char *input, int size)
{
  DESC *k;
  char buf[65], *r=d->input, *s;
  int len;

  /* Copy one line (up to 128 bytes) of input data into the buffer */
  for(s=input;s < input+size && d->input_offset < 128;s++) {
    if(*s == '\r' || *s == '\n')
      break;
    d->input[d->input_offset++] = *s;
  }
  if(!(s < input+size && d->input_offset < 128)) {
    d->timeout=now+3;
    return 1;  /* Grab more input */
  }
    
  d->input[d->input_offset] = '\0';

  /* match client descriptor */
  for(k=descriptor_list;k;k=k->next)
    if((k->flags & CLIENT) && k->local_id == d->local_id)
      break;
  if(!k)
    return 0;  /* Client no longer exists */

  /* parse input string and search for userid */
  for(len=0;len<3;len++,r++)
    if(!(r=strchr(r, ':'))) {
      debug("Ident lookup failed for FD %d: %s\n", k->fd, r);
      return 0;
    }

  while(*r == ' ' || *r == '\t')
    r++;
  if((len=strlen(r)) < 32) {
    debug("Ident lookup succeeded for FD %d: %s\n", k->fd, r);
    if(!*k->hostname)
      memcpy(k->hostname, r, len+1);
    else {
      sprintf(buf, "%s@%.*s", r, 63-len, k->hostname);
      strcpy(k->hostname, buf);
    }
  } else
    debug("Ident lookup failed for FD %d: %s\n", k->fd, r);

  return 0;
}
